# Copyright 2024 Zoé Cassiopée Gauthier.
#
# Use of this source code is governed by an MIT-style
# license that can be found in the LICENSE file or at
# https://opensource.org/licenses/MIT.

from typing import final

from rich.progress import Progress

from pogo_scaled_estimators.pokebattler_proxy import Moveset, PokebattlerProxy, Raid
from pogo_scaled_estimators.utilities import format_pokemon_name


@final
class Calculator:
    def __init__(self, attacker_types: list[str]) -> None:
        self.attacker_types = attacker_types
        self._pokebattler_proxy = PokebattlerProxy()
        self._progress = Progress()

    def calculate(self, level: int = 40, party: int = 1) -> None:
        raid_bosses = self._pokebattler_proxy.raid_bosses(self.attacker_types)
        attackers = {
            attacker: {"RAID_LEVEL_3": [], "RAID_LEVEL_5": [], "RAID_LEVEL_MEGA": []}
            for attacker in self._pokebattler_proxy.with_charged_moves(self.attacker_types)
        }

        with self._progress:
            total_defenders = sum(len(defenders) for defenders in raid_bosses.values())
            task = self._progress.add_task("Working...", total=(total_defenders * 30))

            for raid_tier, defenders in raid_bosses.items():
                if raid_tier == "RAID_LEVEL_MEGA_5":
                    simplified_raid_tier = "RAID_LEVEL_MEGA"
                elif raid_tier == "RAID_LEVEL_ULTRA_BEAST":
                    simplified_raid_tier = "RAID_LEVEL_5"
                else:
                    simplified_raid_tier = raid_tier
                for defender in defenders:
                    self._progress.update(task, description=f"vs {format_pokemon_name(defender)}...")
                    raid = Raid(raid_tier, defender, level, party)
                    results = {
                        attacker: movesets
                        for attacker, movesets in self._pokebattler_proxy.simulate(raid).items()
                        if attacker in attackers
                    }
                    best_movesets = {
                        attacker: min(movesets, key=lambda moveset: moveset.estimator)
                        for attacker, movesets in results.items()
                    }
                    best_estimator = min(best_movesets.values(), key=lambda moveset: moveset.estimator).estimator
                    for attacker, moveset in best_movesets.items():
                        self._progress.update(task, advance=1)
                        attackers[attacker][simplified_raid_tier].append(moveset.scale(best_estimator))

        ase = {}
        for attacker, raid_tier_results in attackers.items():
            if (
                not raid_tier_results["RAID_LEVEL_3"]
                or not raid_tier_results["RAID_LEVEL_5"]
                or not raid_tier_results["RAID_LEVEL_MEGA"]
            ):
                continue
            ase = (
                0.15 * self._average_estimator(raid_tier_results["RAID_LEVEL_3"])
                + 0.50 * self._average_estimator(raid_tier_results["RAID_LEVEL_5"])
                + 0.35 * self._average_estimator(raid_tier_results["RAID_LEVEL_MEGA"])
            )
            print(f"{attacker},{level},{ase}")

    def _average_estimator(self, movesets: list[Moveset]) -> float:
        return sum(moveset.estimator for moveset in movesets) / len(movesets)
